# from qcqp import qcqp
#from __future__ import print_function
# from qcqp import *
import cvxpy as cvx
import numpy as np
from scipy.optimize import root, minimize
#from project_to_ellipsoid import project_to_ellipsoid
from numpy.linalg import norm, inv, slogdet
import scipy.linalg as sla;
from numpy import exp
#import matplotlib.pyplot as plt
import scipy.linalg as sla;
import numpy.random as ra;
import numpy.linalg as la;
import scipy.stats;
import ipdb;
from Functions.objective_functions import LinearModel
from kjunutils import *;
from scipy.special import logsumexp;

'''Place holder class that all GLM bandit algorithms must inherit from GLM '''

def gloc_solve_by_cvx(d,S,theta_prime,At):
    th = cvx.Variable(d)
    obj = cvx.Minimize(cvx.quad_form(th - theta_prime, At))
    cons = [cvx.norm(th) <= S]

    prob = cvx.Problem(obj, cons)
    prob.solve()

    return np.array(th.value).flatten()

def calc_sqrt_beta_det2(d,t,R,ridge,delta,S,logdetV):
  return R * np.sqrt( logdetV - d*log(ridge) + log (1/(delta**2)) ) + sqrt(ridge) * S

def calc_sqrt_beta_thompson(d,t,R,delta):
   return R * np.sqrt(9 * d * np.log(t/delta));

def mu_logistic(z):
    return 1.0/(1.0+np.exp(-z));

def mu_probit(z):
    return scipy.stats.norm.cdf(z);

def logistic_loss_pm1(w, x, y):
    y = float(y);
    assert (y in [+1.0, -1.0]);
    z = y *np.dot(x,w)
    return np.log(1 + np.exp(-z));

def logistic_loss_01(w, x, y):
    assert (y in [0.0, 1.0]);
    yy = y * 2 - 1;
    return logistic_loss_pm1(w,x,yy);

def logistic_loss_pm1_grad(w, x, y):
    y = float(y);
    assert (y in [+1.0, -1.0]);
    z = y * np.dot(x,w);
    return - 1.0 / (1 + np.exp(z)) * (y*x);

def logistic_loss_01_grad(w, x, y):
    assert (y in [0.0, 1.0]);
    yy = y * 2 - 1;
    return logistic_loss_pm1_grad(w,x,yy);

def mfunc_logistic(z):
    return np.log(1.0 + np.exp(z));

def mfunc_logistic_der(z):
    return 1.0/(1.0 + np.exp(-z));

from sklearn import metrics;
def evalAuc(banditObj, dataObj, nTry=1):
    """
        evaluates deployment performance by AUC. 
        repeats are important when the data is synthetic and generated by p(y|x)
    """
    aucList = [];
    trainAucList = [];
    testAucList = [];
    N = dataObj.N;
    for iTry in range(nTry):
        #- 1. generate the labels
        y = np.array([dataObj.get_reward(i) for i in range(dataObj.N)]);
        #- 2. let the bandit compute the scores (0.0-1.0)
        pred = banditObj.predict(dataObj.X);
        #- 3. normalize the pred to be in (0.0-1.0)
        mini = pred.min();
        maxi = pred.max();
        yhat = (pred - mini) / (maxi - mini);
        #- 4. measure auc
        auc = metrics.roc_auc_score(y, yhat)

        assert(np.all(dataObj.X == banditObj.X))
        do_not_ask = banditObj.getDoNotAsk();
        trainAuc = metrics.roc_auc_score(y[do_not_ask], yhat[do_not_ask]);

        testIdx = np.setdiff1d(np.arange(N), do_not_ask);
        testAuc = metrics.roc_auc_score(y[testIdx], yhat[testIdx]);

        aucList.append(auc);
        trainAucList.append(trainAuc);
        testAucList.append(testAuc);

    # print 'debug: check the values'
    # ipdb.set_trace();

    return np.mean(aucList), np.mean(trainAucList), np.mean(testAucList);

################################################################################
# bandit classes
################################################################################
class Bandit(object):
    def __init__(self, X, theta):
        raise NotImplementedError();

    def next_arm(self):
        raise NotImplementedError();

    def update(self):
        raise NotImplementedError();

    def get_debug_dict(self):
        raise NotImplementedError();

class Glm(object):
    def __init__(self):
        raise NotImplementedError();


class GlmLogistic(Glm):
    def __init__(self):
        raise NotImplementedError();

    @staticmethod
    def negloglik(z, y):
        assert (float(y) == 1.0 or float(y) == 0.0);
        return -z*y + GlmLogistic.m(z);

    @staticmethod
    def negloglik_derivative(z, y):
        assert (float(y) == 1.0 or float(y) == 0.0);
        return -y + GlmLogistic.mu(z);

    @staticmethod
    def mu(z):  # link function 
        return 1.0/(1.0+np.exp(-z));

    @staticmethod
    def m(z):   # integral of mu 
        return logsumexp([np.zeros(z.shape), z],axis=0); #np.log(1.0 + np.exp(z));
#        return np.log(1.0 + np.exp(z));

    @staticmethod
    def getKappa(S):
        return 1.0 / ((1 + exp(S)) * (1 + exp(-S)));

class GlmProbit(Glm):
    def __init__(self):
        raise NotImplementedError();

    @staticmethod
    def negloglik(z, y):
        assert (float(y) == 1.0 or float(y) == 0.0);
        return -z*y + GlmProbit.m(z);

    @staticmethod
    def negloglik_derivative(z, y):
        assert (float(y) == 1.0 or float(y) == 0.0);
        return -y + GlmProbit.mu(z);

    @staticmethod
    def mu(z):  # link function
        return scipy.stats.norm.cdf(z);

    @staticmethod
    def m(z):   # integral of mu
        return z*sstats.norm.cdf(z) + sstats.norm.pdf(z)

    @staticmethod
    def getKappa(S):
        return scipy.stats.norm.pdf(S);

########################################
class Gloc(Bandit):
########################################
    """
    R: subgaussian parameter.
    S: norm upper bound of theta_star
    r: range parameter for qgloc
    kappa: lower bound on the dervative of \mu. 
    """    
    def __init__(self, X, lam, R, S, flags, r, kappa, m_t_type, next_arm_method, glm=GlmLogistic, subsample_func=None, subsample_rate=1.0, multiplier=1.0, approx_func=None, m_obj=None, calc_radius_version=2, bArmRemoval=True): #objective, gradient, reward_function):
        self.X = X;
        [self.N, self.d] = X.shape
        self.lam = lam;
        self.R = R;
        self.S = S;
        self.r = r;
        self.kappa = kappa;
        self.multiplier = multiplier;
        self.m_t_type = m_t_type;
        self.next_arm_method = next_arm_method;
        self.glm = glm;

        assert subsample_func == None;
        self.subsample_func = subsample_func;
        self.subN = np.round(self.N * float(subsample_rate)).astype(int);
        self.approx_func = approx_func;
        self.m_obj = m_obj;

        #--- initialize
        #- subsampling aspect

        self.At = np.eye(self.d)*self.lam
        self.invAt = np.eye(self.d)/self.lam
#        theta_s = np.random.normal(0, 1, self.d) # FIXME perhaps change later on.
        # self.theta_s = theta_s / norm(theta_s) * self.S
        self.theta_s = np.zeros(self.d); 
        # print "Note that I am setting theta_s as zeros and this is warned to cause an error in cvx."
        self.theta_hat = self.theta_s.copy();
        #- theta_hat = np.zeros(self.d); # this causes an error in cvx
        self.t = 1
        self.XTz = np.zeros(self.d)
        self.sum_z_t_sq = 0;
        self.radius_problem_constant = 0.;

        self.cvx_th = cvx.Variable(self.d);
        self.cvx_cons = [cvx.norm(self.cvx_th) <= self.S];

        #- do_not_ask_list
        self.do_not_ask = [];
        self.bArmRemoval = bArmRemoval;

        self.dbg_dict = {
            'm_t_type': self.m_t_type,
            'subN': self.subN,
            'multiplier':multiplier,
            'depth_list': [np.nan],
            'lookup_cnt_list': [np.nan],
            'calc_radius_version': calc_radius_version,
        };

    def _calc_radius_sq_v1(self): # the best at around 1.0
        if (self.t == 1):
            radius_sq = 0;
        else:
            radius_sq = self.radius_problem_constant + (self.S**2) * self.lam;
        return radius_sq;

    def _calc_radius_sq_v2(self): # the best at around 1.0
        if (self.t == 1):
            radius_sq = 0;
        else:
            radius_sq = self.radius_problem_constant;
        return radius_sq;

    def _calc_radius_sq_v3(self): # the best at around 0.01
        dt = 0.2; R = self.R;
        if (self.t == 1):
            radius_sq = 0;
        else:
            #- still, omitting some terms.
            B = (0.5/self.kappa)*self.radius_problem_constant + 2*self.kappa*(self.S**2) * self.lam;
            inner = 1 + (2/self.kappa)*B + 4*R**4/(self.kappa**4 * dt**2)
            extra = (self.sum_z_t_sq - self.theta_hat.dot(self.XTz));
            assert (extra > -1e-8);
            alpha = 1 + (4.0/self.kappa)*B \
                      + (8*R**2/self.kappa**2)*np.log((2/dt)*np.sqrt( inner ));
            radius_sq =  alpha\
                         + self.lam*self.S**2 \
                         - extra;
        return radius_sq;

    def _calc_radius_sq(self): 
        v = self.dbg_dict['calc_radius_version'];
        if (v == 1):
            return self._calc_radius_sq_v1();
        elif (v == 2):
            return self._calc_radius_sq_v2();
        elif (v == 3):
            return self._calc_radius_sq_v3();
        else:
            raise ValueError();

    def next_arm(self):
        valid_idx = setdiff1d(np.arange(self.N),self.do_not_ask);

        if (self.t == 1):
            return ra.randint(self.N), np.nan;

        X = self.X;
        theta_hat = self.theta_hat
        At = self.At
        invAt = self.invAt

        radius_sq = self.multiplier * self._calc_radius_sq();
        
        nam = self.next_arm_method;
        if (nam == 'gloc'):
            if (self.subsample_func == None):
                obj_func = np.dot(X, theta_hat)  \
                        + np.sqrt(radius_sq) \
                          * np.sqrt(mahalanobis_norm_sq_batch(X, self.invAt));
#                          * np.sqrt(np.sum(np.dot(X, invAt) * X, axis=1))
                arm_inner = np.argmax(obj_func[valid_idx]);
                arm = valid_idx[arm_inner]
            else:
                raise NotImplementedError(); # todo: use valid_idx...
#                 idx = self.subsample_func(self.N, self.subN);
#                 subX = self.X[idx,:];
#                 obj_func = np.dot(subX, self.theta_hat) + \
#                            + np.sqrt(radius_sq) \
#                              * np.sqrt(mahalanobis_norm_sq_batch(subX, self.invAt));
#                 
#                 arm = idx[np.argmax(obj_func)];
            pass;
        elif (nam == 'qgloc'):
            raise NotImplementedError(); # todo: use validinds..
#             #- compute the smallest eigen value
#             if (self.m_t_type == 'min_sqrt_eig'):
#                 min_ev = sla.eigh(self.invAt, eigvals=(0,0), eigvals_only=True)[0];
#                 m_t = self.r * np.sqrt(min_ev);
#                 assert (m_t + 1e-8 >= self.r / np.sqrt(self.lam + self.t-1));
#             elif (self.m_t_type == 'min_sqrt_eig_fast'):
#                 m_t = self.r / np.sqrt(self.lam + (self.t-1));
#             else: raise ValueError();
# 
#             factor = (radius_sq**0.25)/4/m_t;
#             if (self.subsample_func == None):
#                 obj_func = np.dot(X, theta_hat)  \
#                         + (factor * mahalanobis_norm_sq_batch(X, invAt)); # np.sum(np.dot(X, invAt) * X, axis=1)
#                 arm = np.argmax(obj_func) 
#             else:
#                 factor_invAt = factor*invAt;
#                 tup = self.subsample_func(self.theta_hat, factor_invAt, self.N, self.subN); 
#                 if (len(tup) >= 2):
#                     self.dbg_dict['depth_list'].append(tup[1]);
#                     self.dbg_dict['lookup_cnt_list'].append(tup[2]['lookupCnt']); 
#                 idx = tup[0];
#                 subX = self.X[idx,:];
# 
#                 if (self.approx_func == None):
#                     obj_func = np.dot(subX, self.theta_hat) + factor * mahalanobis_norm_sq_batch(subX, invAt);
#                 else:
#                     obj_func = subX.dot(self.theta_hat) + self.approx_func(self.m_obj, subX, factor_invAt);  
#                 arm = idx[np.argmax(obj_func)];
            
        elif (nam == 'glocts'):
            tmp = ra.normal(size=(self.d,));
            matR = la.cholesky(self.invAt).T;
            
            #- do mult^2 since I want to multiply `mult` to invVt.
            theta_til = self.theta_hat + np.sqrt(radius_sq) * np.dot(tmp, matR);

            if (self.subsample_func == None):
                obj_func = np.dot(self.X, theta_til);
                arm_inner = np.argmax(obj_func[valid_idx]);
                arm = valid_idx[arm_inner];
            else:
                raise NotImplementedError(); # todo: use valid_idx...
#                 tup = self.subsample_func(theta_til, self.N, self.subN);
#                 if (len(tup) >= 2):
#                     self.dbg_dict['depth_list'].append(tup[1]);
#                     self.dbg_dict['lookup_cnt_list'].append(tup[2]['lookupCnt']);
#                 idx = tup[0];
#                 subX = self.X[idx,:]; 
#                 obj_func = np.dot(subX, theta_til);
#                 arm = idx[np.argmax(obj_func)]
        else:
            raise ValueError();
        
        return arm, radius_sq

    def _solve_by_cvx_old(self, theta_prime, At):
        """
        if you want to improve the speed, might want to use quadprog. see
        https://scaron.info/blog/quadratic-programming-in-python.html
        """
        d = self.d
        S = self.S
        th = self.cvx_th;
        cons = self.cvx_cons;
        obj = cvx.Minimize(cvx.quad_form(th - theta_prime, At))

        prob = cvx.Problem(obj, cons)
        prob.solve(); # or ECOS 

        sol = np.array(th.value).flatten();
        assert sol[0] is not None;
        return sol;

    def _solve_by_cvx(self, theta_prime, At):
        """
        if you want to improve the speed, might want to use quadprog. see
        https://scaron.info/blog/quadratic-programming-in-python.html
        """
        d = self.d
        S = self.S
        th = self.cvx_th;
        cons = self.cvx_cons;
        obj = cvx.Minimize(cvx.quad_form(th - theta_prime, At))

        prob = cvx.Problem(obj, cons)
        try: 
            prob.solve();
        except Exception as inst:
            print '#'*40
            print '# ',; print(inst)
            print '#'*40
            print 'try again, with a different solver';
            try:
                prob.solve(solver=cvx.SCS)
            except Exception as inst2:
                ipdb.set_trace();
                pass
#        prob.solve(solver=cvx.SCS); # or ECOS 
#        prob.solve(solver=cvx.SCS, warm_start=True); # or ECOS 
#        prob.solve(warm_start=True, abstol=1e-4, reltol=1e-4, feastol=1e-4); # or ECOS 
#        prob.solve(); # or ECOS 
#        prob.solve(solver=cvx.CVXOPT); # or ECOS 
#        ipdb.set_trace();

        sol = np.array(th.value).flatten();
        assert sol[0] is not None;
        return sol;

    def _solve_by_scipy(self, theta_prime, At):
        """ use the dogleg method! 
        """
        raise NotImplementedError();
#         optimize.minimize(
#                 ..
        pass

    def update(self, pulled_idx, y_t):
        assert (y_t >= 0.0 and y_t <= 1.0);
        X = self.X
        At = self.At
        invAt = self.invAt
        theta_s = self.theta_s
        kappa = self.kappa
        d = self.d
        S = self.S

        x_t = X[pulled_idx, :]
        At_new = At + np.outer(x_t, x_t)
        At_new = .5*(At_new + At_new.T);
        # eigvalsh = sla.eigvalsh(At_new);
        # assert min(eigvalsh) >= 0 and eigvalsh.dtype != complex
        invAt_new = inv(At_new)

        z = np.dot(theta_s, x_t);
        g_t = self.glm.negloglik_derivative(z, y_t); #  (1 / (1 + np.exp(-np.dot(theta_s, x_t))) - y_t)
        theta_prime = theta_s - g_t * np.dot(invAt_new, x_t) / kappa

        try:    
            theta_s = self._solve_by_cvx(theta_prime, At_new)
        except Exception as inst:
            eps = 1e-7;
            print '#'*40
            print '# ',; print(inst)
            print '#'*40
            print 'try again... by adding eps=%g'%eps

            try:
                theta_s = self._solve_by_cvx(theta_prime, At_new + eps*np.eye(len(theta_prime)))
            except Exception as inst2:
                ipdb.set_trace();

        assert theta_s[0] is not None;

        self.At = At_new
        self.invAt = invAt_new
        self.theta_s = theta_s

        #- extra
        self.XTz += np.dot(theta_s, x_t) * x_t
        z = np.dot(theta_s, x_t);
        self.sum_z_t_sq += z**2;
        invAt = self.invAt
        self.radius_problem_constant += (g_t ** 2) * np.dot(x_t, np.dot(invAt, x_t)); # no need for this, but let's keep it this way.
#        self.radius_problem_constant += (g_t ** 2) * np.dot(x_t, np.dot(invAt, x_t)) / (2*kappa) # no need for this, but let's keep it this way.
        self.theta_hat = np.dot(self.invAt, self.XTz)
        if (self.bArmRemoval):
            self.do_not_ask.append( pulled_idx );

        self.t += 1;
    
    def getDoNotAsk(self):
        return self.do_not_ask;

    def predict(self, X=None):
        if X is None:
            X = self.X;
        return X.dot(self.theta_hat)

    def get_debug_dict(self):
        return self.dbg_dict;


########################################
class Oful(Bandit):
########################################
    def __init__(self, X, lam, R, S, flags, subsample_func=None, subsample_rate=1.0, multiplier=1.0):
        self.X = X
        self.R = R;
        self.lam = lam;
        self.delta = .2;
        self.S = S;
        self.flags = flags;
        self.multiplier = float(multiplier);

        # more instance variables
        self.t = 1;
        self.N, self.d = self.X.shape;

        #- subsampling aspect
        assert subsample_func == None;
        self.subN = np.round(self.N * float(subsample_rate)).astype(int);
        self.subsample_func = subsample_func;

        self.XTy = np.zeros(self.d)
        self.invVt = np.eye(self.d) / self.lam
        self.X_invVt_norm_sq = np.sum(self.X * self.X, axis=1) / self.lam;
        self.logdetV = self.d*log(self.lam);
        self.sqrt_beta = calc_sqrt_beta_det2(self.d,self.t,self.R,self.lam,self.delta,self.S,self.logdetV);      
        self.theta_hat = np.zeros(self.d);
        self.Vt = self.lam * np.eye(self.d); 

        self.do_not_ask = [];
        self.dbg_dict = {'multiplier':float(multiplier), 
                'subN': self.subN,
                'subsample_func': self.subsample_func};

    def next_arm(self):
        valid_idx = setdiff1d(np.arange(self.N),self.do_not_ask);
        if (self.t == 1):
            return ra.randint(self.N), np.nan;
        my_c = self.dbg_dict['multiplier']
        radius_sq = self.multiplier * (self.sqrt_beta)**2;
        if (self.subsample_func == None):
            obj_func = np.dot(self.X, self.theta_hat) + np.sqrt(radius_sq) * np.sqrt(self.X_invVt_norm_sq) 
            chosen_inner = np.argmax(obj_func[valid_idx]);
            chosen = valid_idx[chosen_inner];
        else:
            raise NotImplementedError(); # todo: use valid_idx 
#             idx = self.subsample_func(self.N, self.subN);
#             subX = self.X[idx,:];
#             obj_func = np.dot(subX, self.theta_hat) + np.sqrt(radius_sq) * \
#                     np.sqrt( mahalanobis_norm_sq_batch(subX, self.invVt));
#             
#             chosen = idx[np.argmax(obj_func)];
        return chosen, radius_sq;

    def update(self, pulled_idx, y_t):
        assert (y_t >= 0.0 and y_t <= 1.0);
        ##########################
        ## DEBUGGING
        ##########################
        # y_t = 2*y_t - 1;
        xt = self.X[pulled_idx, :]

        self.XTy += (2*y_t - 1) * xt
        self.Vt += np.outer(xt,xt);

        tempval1 = np.dot(self.invVt, xt)    # d by 1, O(d^2)
        tempval2 = np.dot(tempval1, xt)      # scalar, O(d)
        self.logdetV += log(1 + tempval2);

        if (self.t % 20 == 0):
            self.invVt = la.inv(self.Vt); 
            # FIXME: would this be a concern for the simulation?
            if (self.subsample_func == None):
    #            self.X_invVt_norm_sq = mahalanobis_norm_sq_batch(self.X, self.invVt);  # O(Nd^2) 
                self.X_invVt_norm_sq -= (np.dot(self.X, tempval1) ** 2) / (1 + tempval2) # efficient update, O(Nd)
        else:
            self.invVt -= np.outer(tempval1, tempval1) / (1 + tempval2) 
            if (self.subsample_func == None):
                self.X_invVt_norm_sq -= (np.dot(self.X, tempval1) ** 2) / (1 + tempval2) # efficient update, O(Nd)

        self.theta_hat = np.dot(self.invVt, self.XTy)
        self.do_not_ask.append( pulled_idx );

        my_t = self.t + 1
        self.sqrt_beta = calc_sqrt_beta_det2(self.d,my_t,self.R,self.lam,self.delta,self.S,self.logdetV);      

        self.t += 1;

    def getDoNotAsk(self):
        return self.do_not_ask;

    def predict(self, X=None):
        if X is None:
            X = self.X;
        return X.dot(self.theta_hat)

    def get_debug_dict(self):
        return self.dbg_dict;

########################################
class Lts(Bandit):
########################################
    def __init__(self, X, R, flags, subsample_func=None, subsample_rate=1.0,multiplier=1.0):
        self.X = X
        self.R = R;
        self.delta = .2;
        self.flags = flags;
        self.multiplier = multiplier;

        # more instance variables
        self.t = 1;
        self.N, self.d = self.X.shape;

        #- subsampling aspect
        self.subsample_func = subsample_func;
        self.subN = np.round(self.N * float(subsample_rate)).astype(int);

        #- save debug variables
        self.do_not_ask = [];
        assert subsample_func == None, "not implemented"
        self.dbg_dict = {
            'subN':             self.subN,
            'do_chol_downdate': False,
            'use_oful_radius':  False,
            'depth_list':       [np.nan],
            'lookup_cnt_list':  [np.nan],
            'multiplier':       multiplier,
        };
        if (self.dbg_dict['do_chol_downdate']):
            print 'WARNING: do_chol_downdate == True';
        if (self.dbg_dict['use_oful_radius']):
            print 'WARNING: use_oful_radius == True';

        self.logdetV = self.d*log(1.0); # for 'use_oful_radius'

        self.XTy = np.zeros(self.d)
        self.invVt = np.eye(self.d)
        self.matR = np.eye(self.d);
        if (self.dbg_dict['use_oful_radius']):
            S_hat = 1.0; # not ideal.. but a makeshift
            self.sqrt_beta = calc_sqrt_beta_det2(self.d,self.t,self.R,1.0,self.delta,S_hat,self.logdetV);      
        else:
            self.sqrt_beta = calc_sqrt_beta_thompson(self.d,self.t,self.R,self.delta);
        self.theta_hat = np.zeros(self.d);

        self.Vt = np.eye(self.d);
        pass

    def next_arm(self):
        valid_idx = setdiff1d(np.arange(self.N),self.do_not_ask);
        #- 1. sample
        #- if z is Normal(0,I), then R^T * z follows N(0, R^T R)
        #- dot(tmp, self.matR) == dot(self.matR.T, tmp)
        radius_sq = self.multiplier * (self.sqrt_beta ** 2);
        tmp = ra.normal(size=(self.d,));
        theta_til = np.sqrt(radius_sq) * dot(tmp, self.matR) + self.theta_hat;

        #2. choose
        if (self.subsample_func == None):
            obj_func = np.dot(self.X, theta_til);
            chosen_inner = np.argmax(obj_func[valid_idx])
            chosen = valid_idx[chosen_inner];
        else:
            raise NotImplementedError(); # todo: use valid_idx  
#             tup = self.subsample_func(theta_til, self.N, self.subN);
#             if (len(tup) >= 2):
#                 self.dbg_dict['depth_list'].append(tup[1]);
#                 self.dbg_dict['lookup_cnt_list'].append(tup[2]['lookupCnt']);
#             idx = tup[0];
#             subX = self.X[idx,:]; 
#             obj_func = np.dot(subX, theta_til);
#             chosen = idx[np.argmax(obj_func)]

        return chosen, radius_sq;

    def update(self, pulled_idx, y_t):
        assert (y_t >= 0.0 and y_t <= 1.0);
        X = self.X
        xt = X[pulled_idx, :]

        self.XTy += (2*y_t-1) * xt

        self.Vt += np.outer(xt,xt);
        tempval1 = np.dot(self.invVt, xt)
        tempval2 = np.dot(tempval1, xt)
        self.logdetV += log(1 + tempval2);
        if (self.t % 20 == 0):
            # every 20 iterations, compute invVt afresh; maintain Vt then take the inverse.
            self.invVt = la.inv(self.Vt);
        else:
            self.invVt -= np.outer(tempval1, tempval1) / (1 + tempval2) 

        self.theta_hat = np.dot(self.invVt, self.XTy)
        self.do_not_ask.append( pulled_idx );

        # avoid performing downdate, just perform the decomposition.
        if (self.dbg_dict['do_chol_downdate']):
            choldowndate(self.matR, tempval1 / np.sqrt(1 + tempval2)) 
        else:
            self.matR = linalg.cholesky(self.invVt).T;

        self.t += 1
        if (self.dbg_dict['use_oful_radius']):
            S_hat = 1.0; # not ideal.. but a makeshift
            self.sqrt_beta = calc_sqrt_beta_det2(self.d,self.t,self.R,1.0,self.delta,S_hat,self.logdetV);      
        else:
            self.sqrt_beta = calc_sqrt_beta_thompson(self.d,self.t,self.R,self.delta);
        return;

    def getDoNotAsk(self):
        return self.do_not_ask;

    def predict(self, X=None):
        if X is None:
            X = self.X;
        return X.dot(self.theta_hat)

    def get_debug_dict(self):
        return self.dbg_dict;

########################################
class Ol2m(Bandit):
    """
       S is `R` in the paper.
    """
    def __init__(self, X, lam, S, flags={}, multiplier=1.0, calc_radius_version=1, bArmRemoval=True):
        self.X = X;
        [self.N, self.d] = X.shape
        self.lam = float(lam);
        self.S = S;
        self.beta = 1.0/(2*(1 + np.exp(S)));
        self.multiplier = multiplier;

        self.Zt = np.eye(self.d)*self.lam
        self.invZt = np.eye(self.d)/self.lam;
        self.logdetZ = self.d*log(self.lam);
        self.w_t = np.zeros(self.d); # this is theta_hat in the OFUL analogy.
        self.t = 1;

        self.cvx_w = cvx.Variable(self.d);
        self.cvx_cons = [cvx.norm(self.cvx_w) <= self.S]

        self.flags = flags;
        self.do_not_ask = [];
        self.bArmRemoval = bArmRemoval;

        self.dbg_dict = {
            'calc_radius_version': calc_radius_version,
        };

    def _calc_radius_sq_v1(self):
        """
        here, t is at least 2.
        """
        R = self.S; # R is the paper notation
        beta = self.beta;
        dt = .2;
        t = self.t;
        assert (t >= 1);

        tau = np.log( np.ceil(2*np.log2(t)) * 2*(t**2) / dt ); # should use t-1 in place of t, but that will cause log(0) problem.
        # logdet_term = np.log(la.det(self.Z)) - self.d*np.log(self.lam);
        logdet_term = self.logdetZ - self.d*np.log(self.lam);
        gam = 8.0*R + (8.0/beta + (16.0/3)*R)*tau + (2.0/beta)*logdet_term + self.lam*(R**2);

        return gam; # gam == "radius squared"

    def _calc_radius_sq_v2(self):
        logdet_term = self.logdetZ - self.d*np.log(self.lam);
        return logdet_term;
#        return (2.0/beta)*logdet_term + self.lam*(R**2);

    def _calc_radius_sq(self):
        v = self.dbg_dict['calc_radius_version'];
        if (v == 1):
            return self._calc_radius_sq_v1();
        elif (v == 2):
            return self._calc_radius_sq_v2();
        else:
            raise ValueError();
        pass

    def next_arm(self):
        """
        return: (arm index, radius squared)
        """
        valid_idx = setdiff1d(np.arange(self.N), self.do_not_ask);

        if (self.t == 1):
            return ra.randint(self.N), np.nan;

        X = self.X;
        w_t = self.w_t;
        Zt = self.Zt;
        invZt = self.invZt;
        
        radius_sq = self.multiplier * self._calc_radius_sq();
        obj_func = np.dot(X, w_t) \
                + np.sqrt(radius_sq) \
                    * np.sqrt(mahalanobis_norm_sq_batch(X, self.invZt));
        arm_inner = np.argmax(obj_func[valid_idx]);
        arm = valid_idx[arm_inner];

        return arm, radius_sq;

    def _solve_by_cvx(self, w_t, Zt_new, grad_f):
        d = self.d
        R = self.S
#        w = cvx.Variable(d)
        w = self.cvx_w;
        cons = self.cvx_cons;
        # obj = cvx.Minimize(cvx.quad_form(w - w_t, Zt_new) + 2*cvx.sum_entries(grad_f * (w - w_t)))
        obj = cvx.Minimize(cvx.quad_form(w - w_t, Zt_new) + 2*cvx.sum_entries(grad_f * w));
#        cons = [cvx.norm(w) <= R]

        prob = cvx.Problem(obj, cons);
        prob.solve();

        return np.array(w.value).flatten();

    def update(self, pulled_idx, y_t):
        assert (y_t >= 0.0 and y_t <= 1.0);
        X = self.X;
        Zt = self.Zt;
        invZt = self.invZt;
        w_t = self.w_t;
        d = self.d;

        x_t = X[pulled_idx, :];
        x_t_2 = np.sqrt(self.beta/2.0) * x_t;

        tempval1 = np.dot(self.invZt, x_t_2);
        tempval2 = np.dot(tempval1, x_t_2);
        self.logdetZ += np.log(1 + tempval2); # debugging successful.

        Zt_new = Zt + np.outer(x_t_2, x_t_2);
        Zt_new = (Zt_new + Zt_new.T)/2.0;
        invZt_new = la.inv(Zt_new);
#        invZt_new = self.invZt - np.outer(tempval1,tempval1) / (1+tempval2);

        #- prepare to solve the cvx
        grad_f = logistic_loss_01_grad(w_t, x_t, y_t); # debugging successful.

        try:
            self.w_t = self._solve_by_cvx(w_t, Zt_new, grad_f);
        except Exception as inst:
            eps = 1e-9;
            print '#'*40
            print '# ',; print(inst)
            print '#'*40
            print 'try again... by adding eps=%g'%eps
            self.w_t = self._solve_by_cvx(w_t, Zt_new + eps*np.eye(len(w_t)), grad_f);

        self.Zt = Zt_new;
        self.invZt = invZt_new;

        if (self.bArmRemoval):
            self.do_not_ask.append( pulled_idx );
        self.t += 1;

    def getDoNotAsk(self):
        return self.do_not_ask;

    def predict(self, X=None):
        if X is None:
            X = self.X;
        return X.dot(self.w_t)

    def get_debug_dict(self):
        return self.dbg_dict;

########################################
class Glmucb(Bandit):
    def __init__(self, X, lam, R, S, kappa, flags={}, multiplier=1.0, glm=GlmLogistic, bArmRemoval=True):
        self.X = X;
        [self.N, self.d] = X.shape
        self.lam = lam;
        self.R = R;
        self.S = S;
        self.kappa = kappa;
        self.glm = glm;
        self.multiplier = multiplier;

        self.Vt = np.eye(self.d)*self.lam
        self.invVt = np.eye(self.d)/self.lam
        self.theta_hat = np.zeros(self.d); 
        self.t = 1
        self.XTy = np.zeros(self.d)

        #- do_not_ask_list
        self.do_not_ask = [];
        self.history = [];
        self.bArmRemoval = bArmRemoval;

        self.dbg_dict = {
        };

    # def _compute_mahalanobis_distance(self, features, psd_matrix):
        # return np.sqrt(np.sum(np.dot(features, psd_matrix) * features, axis=1))

    def _calc_radius_sq_old(self):
        dt = .2;
        d = self.d;
        alpha = self.R/self.kappa * np.sqrt( .5*self.d*np.log(1 + (2.0*self.t)/d) + np.log(1/dt) );
        return alpha**2;

    def _calc_radius_sq(self):
        dt = .2;
        d = self.d;
        t = self.t;
        k_mu = 0.25; # Lipschitz; would be 0.25 for the logit model. let's just set 0.25 for simplicity.
        c_mu = self.kappa;
        c_m = 1.0;   # norm of the arms. 
        R_max = 1.0; # upper bound on the reward.
        glmucb_kappa = np.sqrt(3 + 2 * np.log(1+2*c_m**2 / self.lam))
        alpha = 2.0*k_mu*glmucb_kappa*R_max/(c_mu) \
                   * np.sqrt( .5*d*np.log(t)*np.log(2*d*t/dt) );
        return alpha**2;

    def next_arm(self):
        valid_idx = setdiff1d(np.arange(self.N), self.do_not_ask);
        if self.t == 1:
            return ra.randint(self.N), np.nan;
    
        X = self.X;
        theta_hat = self.theta_hat
        Vt = self.Vt
        invVt = self.invVt

        radius_sq = self.multiplier * self._calc_radius_sq();
        obj_func = self.glm.mu(np.dot(X, theta_hat))  \
                + np.sqrt(radius_sq) \
                    * np.sqrt(mahalanobis_norm_sq_batch(X, self.invVt));
        arm_inner = np.argmax(obj_func[valid_idx]);
        arm = valid_idx[arm_inner]
        return arm, radius_sq

    def _solve_batch(self):
        #- prepare the objective function
        assert len(self.history) == self.t;
        if self.bArmRemoval:
            assert len(self.do_not_ask) == self.t;
        XTy = self.XTy; 
        pulledX = self.X[self.history,:];
        S = self.S;
        def negloglik(th):
#            return -XTy.dot(th) + np.sum( mfunc_logistic(pulledX.dot(th)) ); 
            return -XTy.dot(th) + np.sum( self.glm.m(pulledX.dot(th)) ); # mfunc_logistic(pulledX.dot(th)) ); 
        def negloglik_grad(th):
#            return -XTy + pulledX.T.dot( mfunc_logistic_der(pulledX.dot(th)) );
            return -XTy + pulledX.T.dot( self.glm.mu(pulledX.dot(th)) );

        cons = {'type': 'ineq', 'fun': lambda th: S ** 2 - th.dot(th)}
        res = minimize(negloglik, self.theta_hat, jac=negloglik_grad, constraints=cons); 
#        res = minimize(negloglik, np.zeros(self.d), jac=negloglik_grad, constraints=cons); 
        theta_hat = np.array(res['x']).flatten();

        return theta_hat;

    def update(self, pulled_idx, y_t):
        assert (y_t >= 0.0 and y_t <= 1.0);
        X = self.X
        Vt = self.Vt
        invVt = self.invVt
        theta_hat = self.theta_hat
        kappa = self.kappa
        d = self.d
        S = self.S

        x_t = X[pulled_idx, :];
        Vt_new = Vt + np.outer(x_t, x_t);
        Vt_new = .5*(Vt_new + Vt_new.T);
        invVt_new = inv(Vt_new);

        self.Vt = Vt_new;
        self.invVt = invVt_new;
        self.XTy += x_t * y_t;
        self.history.append( pulled_idx );
        if (self.bArmRemoval):
            self.do_not_ask.append( pulled_idx );

        #- run the optimization
        self.theta_hat = self._solve_batch();

        self.t += 1;

    def getDoNotAsk(self):
        return self.do_not_ask;

    def predict(self, X=None):
        if X is None:
            X = self.X;
        return X.dot(self.theta_hat)
    
    def get_debug_dict(self):
        return self.dbg_dict;

########################################
class GlmucbLi(Bandit):
    def __init__(self, X, lam, R, S, kappa, flags={}, multiplier=1.0, glm=GlmLogistic, bArmRemoval=True):
        self.X = X;
        [self.N, self.d] = X.shape
        self.lam = lam;
        self.R = R;
        self.S = S;
        self.kappa = kappa;
        self.glm = glm;
        self.multiplier = multiplier;

        self.Vt = np.eye(self.d)*self.lam
        self.invVt = np.eye(self.d)/self.lam
        self.theta_hat = np.zeros(self.d); 
        self.t = 1
        self.XTy = np.zeros(self.d)

        #- do_not_ask_list
        self.do_not_ask = [];
        self.history = [];
        self.bArmRemoval = bArmRemoval;

        self.dbg_dict = {
        };

    # def _compute_mahalanobis_distance(self, features, psd_matrix):
        # return np.sqrt(np.sum(np.dot(features, psd_matrix) * features, axis=1))

    def _calc_radius_sq(self):
        dt = .2;
        d = self.d;
        t = self.t;
        alpha = self.R/self.kappa * np.sqrt(d/2.0*np.log(1+2.0*t/d) + np.log(1.0/dt));
        
        return alpha**2;

    def next_arm(self):
        valid_idx = setdiff1d(np.arange(self.N), self.do_not_ask);
        if self.t == 1:
            return ra.randint(self.N), np.nan;
    
        X = self.X;
        theta_hat = self.theta_hat
        Vt = self.Vt
        invVt = self.invVt

        radius_sq = self.multiplier * self._calc_radius_sq();
        obj_func = np.dot(X, theta_hat)  \
                + np.sqrt(radius_sq) \
                    * np.sqrt(mahalanobis_norm_sq_batch(X, self.invVt));
        arm_inner = np.argmax(obj_func[valid_idx]);
        arm = valid_idx[arm_inner]
        return arm, radius_sq

    def _solve_batch(self):
        #- prepare the objective function
        assert len(self.history) == self.t;
        if self.bArmRemoval:
            assert len(self.do_not_ask) == self.t;
        XTy = self.XTy; 
        pulledX = self.X[self.history,:];
        S = self.S;
        def negloglik(th):
#            return -XTy.dot(th) + np.sum( mfunc_logistic(pulledX.dot(th)) ); 
            return -XTy.dot(th) + np.sum( self.glm.m(pulledX.dot(th)) ); # mfunc_logistic(pulledX.dot(th)) ); 
        def negloglik_grad(th):
#            return -XTy + pulledX.T.dot( mfunc_logistic_der(pulledX.dot(th)) );
            return -XTy + pulledX.T.dot( self.glm.mu(pulledX.dot(th)) );

        cons = {'type': 'ineq', 'fun': lambda th: S ** 2 - th.dot(th)}
        res = minimize(negloglik, self.theta_hat, jac=negloglik_grad, constraints=cons); 
#        res = minimize(negloglik, np.zeros(self.d), jac=negloglik_grad, constraints=cons); 
        theta_hat = np.array(res['x']).flatten();

        return theta_hat;

    def update(self, pulled_idx, y_t):
        assert (y_t >= 0.0 and y_t <= 1.0);
        X = self.X
        Vt = self.Vt
        invVt = self.invVt
        theta_hat = self.theta_hat
        kappa = self.kappa
        d = self.d
        S = self.S

        x_t = X[pulled_idx, :];
        Vt_new = Vt + np.outer(x_t, x_t);
        Vt_new = .5*(Vt_new + Vt_new.T);
        invVt_new = inv(Vt_new);

        self.Vt = Vt_new;
        self.invVt = invVt_new;
        self.XTy += x_t * y_t;
        self.history.append( pulled_idx );
        if (self.bArmRemoval):
            self.do_not_ask.append( pulled_idx );

        #- run the optimization
        self.theta_hat = self._solve_batch();

        self.t += 1;

    def getDoNotAsk(self):
        return self.do_not_ask;

    def predict(self, X=None):
        if X is None:
            X = self.X;
        return X.dot(self.theta_hat)
    
    def get_debug_dict(self):
        return self.dbg_dict;

########################################
class Uniform(Bandit):
    def __init__(self, X, flags={}):
        self.X = X;
        [self.N, self.d] = X.shape
        self.flags = flags;
        self.do_not_ask = [];
        self.dbg_dict = {};
        self.t = 1;
    
    def next_arm(self):
        valid_idx = setdiff1d(np.arange(self.N),self.do_not_ask);
        return ra.choice(valid_idx), np.nan;

    def update(self, pulled_idx, y_t):
        self.do_not_ask.append( pulled_idx );
        self.t += 1;
        pass
    
    def getDoNotAsk(self):
        return self.do_not_ask;

    def predict(self, X=None):
        if X is None:
            X = self.X;
        return ra.normal(size=X.shape[0]);

    def get_debug_dict(self):
        return self.dbg_dict;

########################################
class NearestNeighbor(Bandit):
    """
    find the closest to the first positive item;
    internal: for each positive point, I maintain `dist_sq` containing the distance_squared  
              to all data points.
    """
    def __init__(self, X, bAdaptive, flags={}):
        self.X = X;
        [self.N, self.d] = X.shape
        self.bAdaptive = bAdaptive;
        self.flags = flags;
        self.do_not_ask = [];
        self.posIdxList = [];
        self.dist_sq_list = [];
        self.dbg_dict = {};
        self.t = 1;
    
    def next_arm(self):
        if (not self.bAdaptive):
            assert len(self.do_not_ask) != 0;

        valid_idx = setdiff1d(np.arange(self.N),self.do_not_ask);
        # seenPositive = (self.x_first_positive is not None);
        if len(self.posIdxList) == 0:
            chosen_final = ra.choice(valid_idx);
        else:
            # if (not seenPositive):
            #     chosen = ra.choice(valid_idx);
            # else:
            if (not self.bAdaptive or len(self.dist_sq_list) == 0):
                #- finding the closest to the query 
                dist_sq = self.dist_sq_list[0][valid_idx];

                chosen_inner = np.argmin(dist_sq);
                chosen_final = valid_idx[chosen_inner];
            else:
                #- consider every positive example encountered so far
                minValList = [];
                argminList = [];
                for dist_sq in self.dist_sq_list:
                    dist_sq = dist_sq[valid_idx];

                    chosen_inner = np.argmin(dist_sq);
                    chosen = valid_idx[chosen_inner];

                    minValList.append( dist_sq[chosen_inner] );
                    argminList.append( chosen );

                minIdx = np.argmin(minValList);
                chosen_final = argminList[minIdx];

        return chosen_final, np.nan;

    def update(self, pulled_idx, y_t):
        self.do_not_ask.append( pulled_idx );

        if (y_t == 1.0): 
            self.posIdxList.append( pulled_idx );

            #- update the distance squared list
            x = self.X[pulled_idx,:];
            dist_sq = ((self.X - x)**2).sum(1); # distance_squared to the rest of the data
            self.dist_sq_list.append(dist_sq);

        self.t += 1;
        pass

    def getDoNotAsk(self):
        return self.do_not_ask;

    def predict(self, X=None):
        if X is None:
            X = self.X;
        assert np.all(X == self.X);
        assert len(self.dist_sq_list) == len(self.posIdxList);

        N = X.shape[0];
        posX = X[self.posIdxList];
        scoreList = [];
        for i in range(X.shape[0]):
            x = X[i,:];

            #- find the closest posIdx
            distances = [];
            for j in range(len(self.posIdxList)):
                posIdx = self.posIdxList[j];
                dist_sq = self.dist_sq_list[j];
                distances.append( dist_sq[i] );

            scoreList.append( np.min(distances) );

        scoreList = -np.array(scoreList); # take negative!
        return scoreList;
    
    def get_debug_dict(self):
        return self.dbg_dict;

########################################
#- a set of exploration heuristics to be used.

import sklearn.linear_model as lm;
import sklearn.svm as svm

def exploreNone(obj_func, valid_idx):
    chosen_inner = np.argmax( obj_func[valid_idx] );
    return valid_idx[chosen_inner];

def exploreEpsGreedy(obj_func, valid_idx, p):
    if (ra.rand() < p):
        chosen = ra.choice(valid_idx);
    else:
        chosen = exploreNone(obj_func, valid_idx);
    return chosen;

def exploreUncertainty(obj_func, valid_idx, t, tThreshold):
    if (t <= tThreshold):
        chosen_inner = np.argmin( np.abs(obj_func[valid_idx]) );
        chosen = valid_idx[chosen_inner];
    else:
        chosen = exploreNone(obj_func, valid_idx);
    return chosen;

########################################
# pure-exploitation
class ScikitBase(Bandit):
    def __init__(self, algo, paramTup, X, exploreAlgo='none', exploreParamTup=(), flags={}):
        self.X = X
        self.flags = flags;

        # more instance variables
        self.t = 1;
        self.N, self.d = self.X.shape;

        self.dbg_dict = { # add debug info
        };

        self.pulledList = [];
        self.yList = [];
        self.do_not_ask = [];
        if (algo == 'ridge'):
            self.model = lm.RidgeClassifier(alpha=paramTup[0], fit_intercept=False); #LogisticRegression(C=paramTup[0], fit_intercept=False);
        elif (algo == 'lr'):
            self.model = lm.LogisticRegression(C=paramTup[0], fit_intercept=False);
        elif (algo == 'svm'):
            #- NOTE that there is no way to remove the intercept for SVM
            self.model = svm.SVC(C=paramTup[0], kernel='linear');
        else:
            raise ValueError();
        self.exploreAlgo = exploreAlgo;
        self.exploreParamTup = exploreParamTup;
        pass

    def _trainModelSafe(self, myX, myY):
        myY = np.array(myY);
        if (np.all(myY == -1) or np.all(myY == 1)):
            seenLabel = myY[0];
            tmp_x = np.zeros(self.d);
            tmp_y = 1.0 - seenLabel;
            tmpX = np.vstack( (myX, tmp_x) );
            tmpY = np.concatenate( (myY, [tmp_y]) );
        else:
            tmpX = myX; tmpY = myY;
        self.model.fit(tmpX, tmpY);

    def next_arm(self):
        if self.t == 1:
            return ra.randint(self.N), np.nan;

        valid_idx = setdiff1d(np.arange(self.N),self.do_not_ask);
        radius_sq = 0.;

        #- train model (prevents from ValueError with 'no 2 classes'
        self._trainModelSafe(self.X[self.pulledList,:], self.yList);

        #- choose an arm
        obj_func = self.model.decision_function(self.X);
        if self.exploreAlgo == 'none':
            chosen = exploreNone(obj_func, valid_idx);
        elif self.exploreAlgo == 'epsgreedy':
            chosen = exploreEpsGreedy(obj_func, valid_idx, self.exploreParamTup[0]);
        elif self.exploreAlgo == 'uncertainty':
            chosen = exploreUncertainty(obj_func, valid_idx, self.t, self.exploreParamTup[0]);
        else: 
            raise ValueError();

        return chosen, radius_sq;

    def update(self, pulled_idx, y_t):
        assert (y_t >= 0.0 and y_t <= 1.0);
        self.pulledList.append(pulled_idx);
        self.yList.append(2*y_t-1);
        self.do_not_ask.append(pulled_idx);

        self.t += 1
        pass

    
    def getDoNotAsk(self):
        return self.do_not_ask;

    def predict(self, X=None):
        if X is None:
            X = self.X;
        pred = self.model.decision_function(X)
        return pred;

    def get_debug_dict(self):
        return self.dbg_dict;


################################################################################
# for experiments
################################################################################
class DataForBandit(object):
    def __init__(self):
        raise NotImplementedError();

    def gen_data(self):
        raise NotImplementedError();

    def get_reward(self, idx):
        raise NotImplementedError();

class SphericalGaussian(DataForBandit):
    def __init__(self, R):
        self.R = R;

    def set_X(self, X):
        self.X = X;
        [self.N, self.d] = X.shape;

    def gen_theta_star(self, S=1.0):
        self.S = S;

        #- generate theta_star
        v = ra.normal(0,1,self.d);
        v /= la.norm(v)
        self.theta_star = v * self.S;

        #- save expected rewards
        self.expt_reward = np.dot(self.X, self.theta_star);
        self.best_arm = np.argmax(self.expt_reward);

    def gen_data(self, d, N, S=1.0):
        self.d = d;
        self.N = N;
        self.S = S;

        #- generate theta_star
        v = ra.normal(0,1,self.d);
        v /= la.norm(v)
        self.theta_star = v * self.S;

        #- generate X
        X = ra.normal(0,1,(self.N, self.d));
        norms = la.norm(X, axis=1);
        X /= norms.reshape(-1,1)
        self.X = X;

        #- save expected rewards
        self.expt_reward = np.dot(self.X, self.theta_star);
        self.best_arm = np.argmax(self.expt_reward);

    def get_reward(self, idx):
        x = self.X[idx,:];
        return np.dot(x, self.theta_star) + self.R * ra.normal(0,1)

class SphericalLogistic(DataForBandit):
    def __init__(self):
        self.X = None;
        pass

    def set_X(self, X):
        self.X = X;
        [self.N, self.d] = X.shape;

    def gen_theta_star(self, S=1.0, rand_stream=ra):
        assert self.X is not None;
        self.S = S;
        #- generate theta_star
        v = rand_stream.normal(0,1,self.d);
        v /= la.norm(v)
        self.theta_star = v * self.S;

        #- save the reward generation model
        self.lm = LinearModel(self.theta_star, 0.0);

        #- save expected rewards
        self.expt_reward_linear = np.dot(self.X, self.theta_star);
        self.expt_reward = mu_logistic( self.expt_reward_linear );
#        self.expt_reward = np.dot(self.X, self.theta_star);
        self.best_arm = np.argmax(self.expt_reward);

    def gen_data(self, d, N, S=1.0, rand_stream=ra):
        self.d = d;
        self.N = N;
        self.S = S;

        #- generate X
        X = rand_stream.normal(0,1,(self.N, self.d));
        norms = la.norm(X, axis=1);
        X /= norms.reshape(-1,1)
        self.X = X;

        self.gen_theta_star(S=self.S, rand_stream=rand_stream);

    def get_reward(self, idx):
        x = self.X[idx,:];
        return self.lm.get_binary_reward(x) #returns 0/1

    def get_expected_regret(self, idx):
        return self.expt_reward[self.best_arm] - self.expt_reward[idx];

class SphericalBiasLogistic(DataForBandit):
    """
        now with bias. can set the bias manually as well.
    """
    def __init__(self, ):
        self.X = None;
        self.bias = None;
        self.theta_star = None;
        pass

    def set_X(self, X):
        raise NotImplementedError();
        # self.X = X;
        # [self.N, self.d] = X.shape;

    def gen_theta_star(self, S=1.0, bias='random', rand_stream=ra):
        assert self.X is not None;
        self.S = S;
        #- generate theta_star
        v = rand_stream.normal(0, 1, self.d + 1);
        if (bias != 'random'):
            v[-1] = bias;
        v /= la.norm(v)

        self.bias = bias;
        self.theta_star = v * self.S;

        #- save the reward generation model
        self.lm = LinearModel(self.theta_star, 0.0);

        #- save expected rewards
        self.expt_reward_linear = np.dot(self.X, self.theta_star);
        self.expt_reward = mu_logistic( self.expt_reward_linear );
#        self.expt_reward = np.dot(self.X, self.theta_star);
        self.best_arm = np.argmax(self.expt_reward);

    def gen_data(self, d, N, S=1.0, bias='random', rand_stream=ra):
        self.d = d;
        self.N = N;
        self.S = S;

        #- generate X
        X = rand_stream.normal(0,1,(self.N, self.d));
        norms = la.norm(X, axis=1);
        X /= norms.reshape(-1,1)
        X = np.hstack( (X,np.ones( (N,1) )) ); # bias
        norms = la.norm(X, axis=1);
        X /= norms.reshape(-1,1) # normalize again. 
        self.X = X;

        self.gen_theta_star(S=self.S, bias=bias, rand_stream=rand_stream);

    def get_reward(self, idx):
        x = self.X[idx,:];
        return self.lm.get_binary_reward(x) #returns 0/1

    def get_expected_regret(self, idx):
#        return mu_logit(self.expt_reward[self.best_arm]) - mu_logit(self.expt_reward[idx]);
        return self.expt_reward[self.best_arm] - self.expt_reward[idx];

class SphericalProbit(DataForBandit):
    def __init__(self):
        self.X = None;
        pass

    def set_X(self, X):
        self.X = X;
        [self.N, self.d] = X.shape;

    def gen_theta_star(self, S=1.0, rand_stream=ra):
        assert self.X is not None;
        self.S = S;
        #- generate theta_star
        v = rand_stream.normal(0,1,self.d);
        v /= la.norm(v)
        self.theta_star = v * self.S;

        #- save the reward generation model
#        self.lm = LinearModel(self.theta_star, 0.0);
        self.lm = None; # not sure if this is needed.

        #- save expected rewards
        self.expt_reward_linear = np.dot(self.X, self.theta_star);
        self.expt_reward = mu_probit( self.expt_reward_linear );
#        self.expt_reward = np.dot(self.X, self.theta_star);
        self.best_arm = np.argmax(self.expt_reward);

    def gen_data(self, d, N, S=1.0, rand_stream=ra):
        self.d = d;
        self.N = N;
        self.S = S;

        #- generate X
        X = rand_stream.normal(0,1,(self.N, self.d));
        norms = la.norm(X, axis=1);
        X /= norms.reshape(-1,1)
        self.X = X;

        self.gen_theta_star(S=self.S, rand_stream=rand_stream);

    def get_reward(self, idx):
        x = self.X[idx,:];
        r = float(ra.rand() <= mu_probit( x.dot(self.theta_star) ));
        return r; # returns 0/1

    def get_expected_regret(self, idx):
        return self.expt_reward[self.best_arm] - self.expt_reward[idx];

class SphericalNN(DataForBandit):
    def __init__(self):
        self.X = None;
        pass

    def set_X(self, X):
        self.X = X;
        [self.N, self.d] = X.shape;

    def gen_expt_reward(self, posFraction, rand_stream=ra):
        assert self.X is not None;

        #- generate theta_star
        v = rand_stream.normal(0,1,self.d);
        v /= la.norm(v)
        self.theta_star_0 = v; 

        #- save expected rewards
        self.expt_reward_0 = np.dot(self.X, self.theta_star_0);
        # self.best_arm = np.argmax(self.expt_reward);

        self.posFraction = posFraction;
        nPos = int(np.round(self.N*posFraction));
        self.nPos = nPos;
        assert(nPos >= 1 and nPos < self.N);

        iprod = self.X.dot(self.theta_star_0);
        sidx = np.argsort(iprod);
        sidx = sidx[::-1]; # make it descending.
        self.expt_reward = np.zeros(self.N);
        self.expt_reward[sidx[:self.nPos]] = 1.0;
        self.best_arm = np.argmax(self.expt_reward);

    def gen_data(self, d, N, posFraction=0.5, rand_stream=ra):
        assert posFraction >= 0 and posFraction <= 1.0
        self.d = d;
        self.N = N;

        #- generate X
        X = rand_stream.normal(0,1,(self.N, self.d));
        norms = la.norm(X, axis=1);
        X /= norms.reshape(-1,1)
        self.X = X;

        self.gen_expt_reward(posFraction, rand_stream=ra);

    def get_reward(self, idx):
        return self.expt_reward[idx]; # returns 0/1

    def get_expected_regret(self, idx):
        return self.expt_reward[self.best_arm] - self.expt_reward[idx];

#- data loading
def kjun_load_data(data_name):
    if (data_name == 'sphere-v01'):
        loaded = LoadPickle('./data/sphere-v01.pkl');
        X = loaded['X'];
        idx_ary = loaded['idxAry'];
        data_obj = SphericalLogistic();
        data_obj.set_X(X);
    elif (data_name == 'sphere-v02'):
        loaded = LoadPickle('./data/sphere-v02.pkl');
        X = loaded['X'];
        idx_ary = loaded['idxAry'];
        data_obj = SphericalLogistic();
        data_obj.set_X(X);
    elif (data_name == 'sphere-v03'):
        loaded = LoadPickle('./data/sphere-v03.pkl');
        X = loaded['X'];
        idx_ary = loaded['idxAry'];
        data_obj = SphericalLogistic();
        data_obj.set_X(X);
    elif (data_name == 'sphere-v04'):
        loaded = LoadPickle('./data/sphere-v04.pkl');
        X = loaded['X'];
        idx_ary = loaded['idxAry'];
        data_obj = SphericalLogistic();
        data_obj.set_X(X);
    elif (data_name == 'sphere-v05-small'):
        loaded = LoadPickle('./data/sphere-v05-small.pkl');
        X = loaded['X'];
        idx_ary = loaded['idxAry'];
        data_obj = SphericalLogistic();
        data_obj.set_X(X);
        fname_lsh_quad = './data/sphere-v05-small-index32-b-l24-k12.pkl';
        fname_lsh_nonquad = './data/sphere-v05-small-nonquad-index32-b-l24-k12.pkl';
    elif (data_name == 'sphere-v05'):
        fName = './data/sphere-v05.pkl';
        print 'Loading: '+fName;
        loaded = LoadPickle('./data/sphere-v05.pkl');
        X = loaded['X'];
        idx_ary = loaded['idxAry'];
        data_obj = SphericalLogistic();
        data_obj.set_X(X);
        fname_lsh_quad = './data/sphere-v05-index32-b-l24-k12.pkl';
        fname_lsh_nonquad = './data/sphere-v05-nonquad-index32-b-l24-k12.pkl';
    elif (data_name == 'sphere-v06'):
        fName = './data/sphere-v06.pkl';
        print 'Loading: '+fName;
        loaded = LoadPickle(fName);
        X = loaded['X'];
        idx_ary = loaded['idxAry'];
        data_obj = SphericalLogistic();
        data_obj.set_X(X);
        fname_lsh_quad = './data/sphere-v06-index32-b-l24-k12.pkl';
        fname_lsh_nonquad = './data/sphere-v06-nonquad-index32-b-l24-k12.pkl';
    elif (data_name == 'sphere-v07'):
        fName = './data/sphere-v07.pkl';
        print 'Loading: '+fName;
        loaded = LoadPickle(fName);
        X = loaded['X'];
        idx_ary = loaded['idxAry'];
        data_obj = SphericalLogistic();
        data_obj.set_X(X);
        fname_lsh_quad = './data/sphere-v07-index32-b-l24-k12.pkl';
        fname_lsh_nonquad = './data/sphere-v07-nonquad-index32-b-l24-k12.pkl';
    elif (data_name == 'sphere-v08'):
        fName = './data/%s.pkl' % data_name;
        print 'Loading: '+fName;
        loaded = LoadPickle(fName);
        X = loaded['X'];
        idx_ary = loaded['idxAry'];
        data_obj = SphericalLogistic();
        data_obj.set_X(X);
        fname_lsh_quad = './data/%s-index32-b-l24-k12.pkl'%data_name;
        fname_lsh_nonquad = './data/%s-nonquad-index32-b-l24-k12.pkl'%data_name;
#     elif (data_name == 'sphere-v08-small'):
#         fName = './data/%s.pkl' % data_name;
#         print 'Loading: '+fName;
#         loaded = LoadPickle(fName);
#         X = loaded['X'];
#         idx_ary = loaded['idxAry'];
#         data_obj = SphericalLogistic();
#         data_obj.set_X(X);
#         fname_lsh_quad = './data/%s-index32-b-l24-k12.pkl'%data_name;
#         fname_lsh_nonquad = './data/%s-nonquad-index32-b-l24-k12.pkl'%data_name;
    else:
        raise ValueError();

    return data_obj, idx_ary, fname_lsh_quad, fname_lsh_nonquad;

def run_bandit(bandit,data_obj,T,initIdx=-1,timeList=[]):
    reward_ary = np.zeros(T)
    arm_ary = np.zeros(T, dtype=int)
    inst_regret = np.zeros(T);
    cum_regret = np.zeros(T);

    #- initial point, if given
    if initIdx != -1:
        bandit.update(initIdx, 1);

    my_tt = tic();
    for t in xrange(1,T+1):
        #- choose the next arm
        next_arm, radius_sq = bandit.next_arm()

        #- get reward and update the model
        reward = data_obj.get_reward(next_arm);
        bandit.update(next_arm, reward);

        #- save stats
        reward_ary[t-1] = reward;
        arm_ary[t-1] = next_arm;
        inst_regret[t-1] = data_obj.get_expected_regret(next_arm);
        if (t == 1):
            cum_regret[t-1] = inst_regret[t-1];
        else:
            cum_regret[t-1] = cum_regret[t-2] + inst_regret[t-1];

        #- print out stats
        if (t % 100 == 0):
            timeSoFar = toc(my_tt);
            print('t=%d, time=%.1f, radius_sq= %.4f, inst_reg=%.4f, cum_reg=%.4f' % \
                    (t, timeSoFar, radius_sq, inst_regret[t-1], cum_regret[t-1]));
            timeList.append( [t,timeSoFar] );

    return reward_ary, arm_ary, bandit.get_debug_dict();

def run_bandit_time(bandit,data_obj,T,initIdx=-1):
    timeList = [];
    reward_ary, arm_ary, dbg_dict = run_bandit(bandit,data_obj,T,initIdx, timeList=timeList);
    return reward_ary, arm_ary, dbg_dict, timeList;

################################################################################
# evaluation functions
################################################################################


class EvalRegret(object):
    def __init__(self, data_obj, arm_idx_ary):
        self.data_obj = data_obj;
        self.arm_idx_ary = arm_idx_ary;

    def get_cum_regret(self):
        do = self.data_obj;

        best_reward = np.max(do.expt_reward);
        return np.cumsum(best_reward - do.expt_reward[self.arm_idx_ary]);

    def get_cum_regret_mu(self,mu_func):
        do = self.data_obj;

        best_reward = np.max(do.expt_reward);
        return np.cumsum(mu_func(best_reward) - mu_func(do.expt_reward[self.arm_idx_ary]));

################################################################################
#- hashing wrapper classes
################################################################################

class LshNonquadWrap(object):
    def __init__(self, lsh, lsh_index_array):
        if (lsh is None):
          return; 
        self.lsh = lsh;
        self.index_array = lsh_index_array;

        tmp = np.zeros(len(lsh_index_array),dtype=int);                      
        tmp[self.index_array.tolist()] = range(len(lsh_index_array)); 
        self.index_array_inv = tmp;                     

    def FindUpto(self, query, max_dist_comp, randomize, invalidList, maxLookup=5000):
        myQuery = np.zeros((len(query),1), dtype=float32);
        myQuery[:,0] = query;
        lshInvalidSet = set( [self.index_array_inv[x] for x in invalidList] );
        foundSet, foundListTuple, dbgDict = self.lsh.FindUpto(myQuery, max_dist_comp, 
                                          randomize=randomize, invalidSet=lshInvalidSet, maxLookup=maxLookup)
        maxDepth = max([x[1] for x in foundListTuple])
        foundList = [self.index_array[x[0]] for x in foundListTuple]

        #- when there is not enough things, fill it out with 
        nFound = len(foundList);
        dbgDict['nRetrievedFromLsh'] = nFound;
        N = len(self.index_array);
        if (nFound < max_dist_comp):
          wholeSet = set(xrange(len(self.index_array)));
          #- pick (max_dist_comp - nFound) data points
          remainder = max_dist_comp - nFound;
          foundList += ra.permutation(list(wholeSet - set(invalidList) - set(foundList)))[:remainder].tolist();

        return foundList, maxDepth, dbgDict;

#---
class LshQuadWrap(object):
    def __init__(self, lsh, lsh_index_array):
        if (lsh is None):
          return; 
        self.lsh = lsh;
        self.index_array = lsh_index_array;

        tmp = np.zeros(len(lsh_index_array),dtype=int);                      
        tmp[self.index_array.tolist()] = range(len(lsh_index_array)); 
        self.index_array_inv = tmp;                     

    def FindUpto(self, query_1_vec, query_2_mat, max_dist_comp, randomize, invalidList, maxLookup=5000, approxFunc=None):
        d = len(query_1_vec);
        myQuery = np.zeros((d + d**2,1), dtype=float32);
        myQuery[:d,0] = query_1_vec;
        myQuery[d:,0] = query_2_mat.ravel();
        lshInvalidSet = set( [self.index_array_inv[x] for x in invalidList] );
        foundSet, foundListTuple, dbgDict = self.lsh.FindUpto(myQuery, max_dist_comp, 
                                          randomize=randomize, invalidSet=lshInvalidSet, maxLookup=maxLookup, approxFunc=approxFunc);
        if (len(foundSet) == 0):
            maxDepth = 0;
        else:
            maxDepth = max([x[1] for x in foundListTuple])
        foundList = [self.index_array[x[0]] for x in foundListTuple]

        #- when there is not enough things, fill it out with 
        nFound = len(foundList);
        dbgDict['nRetrievedFromLsh'] = nFound;
        N = len(self.index_array);
        if (nFound < max_dist_comp):
          wholeSet = set(range(len(self.index_array)));
          #- pick (max_dist_comp - nFound) data points
          remainder = max_dist_comp - nFound;
          foundList += ra.permutation(list(wholeSet - set(invalidList) - set(foundList)))[:remainder].tolist();

        return foundList, maxDepth, dbgDict;

################################################################################
# subsample functions
################################################################################
# def subsample_light_slow(N, subN):
#     return ra.permutation(N)[:subN];  

# def subsample_light(N, subN):
#     return ra.randint(N,size=subN);

print "WARNING using slow permutation for light"
def subsample_light(N, subN):
    return ra.permutation(N)[:subN];  

def subsample_qoful_light(theta_hat, mat, N, subN): 
    return (subsample_light(N, subN), );

def subsample_qoful_hash(theta_hat, mat, N, subN):
    raise NotImplementedError();

def subsample_gen_qoful_hash(lsh_wrap, iprod_type='full', m=None, d=None):
    """ iprod_type = ('full' | 'l1' | 'l2') """
    if (iprod_type == 'full'):
        approx_func = None;
    elif (iprod_type == 'l1'):
        approx_func = lambda projections_all, query: iprod.qoful_calc_v3_l1(m, projections_all, query, d)
    elif (iprod_type == 'l2'):
        raise NotImplementedError();
    else:
        raise ValueError();

    def afunc(theta_hat, mat, N, subN):
        """ uses lsh here """
        return lsh_wrap.FindUpto(theta_hat, mat, subN, True, [], approxFunc=approx_func);
    return afunc

def subsample_lts_light(theta_til, N, subN):
    return (subsample_light(N, subN), );

def subsample_gen_lts_hash(lsh_wrap):
    def subsample_lts_hash(theta_til, N, subN):
        return lsh_wrap.FindUpto(theta_til, subN, True, []);
    return subsample_lts_hash
    
